// SPDX-FileCopyrightText: 2025 UnionTech Software Technology Co., Ltd.
//
// SPDX-License-Identifier: LGPL-3.0-or-later

#include "linyaps_box/printer.h"

namespace linyaps_box {

printer::~printer() = default;

} // namespace linyaps_box
